package spads1;

// Class created by Simon Dellicour (june 2011)

import java.util.ArrayList;

public class Amova 
{
	public SimThread simTestThread;
	
	public Amova()
		{	}
	
	// Method to compute the AMOVA F-statistics (Excoffier et al, 1992) of a "groups" ArrayList:
	public double[] amova(int h, Locus locus, ArrayList<Group> groups, ArrayList<Node> nodes, int numberOfPermutations, SimThread simTestThread0)
		{
			double[] Fstats = new double[12];
			double[] Fstats0 = new double[9];
		    double pvalFsc = 0; double pvalFst = 0; double pvalFct = 0;
			simTestThread = simTestThread0;
			
			Fstats0 = this.amovaFstatistics(locus, groups, nodes, false, false);
			
			if (numberOfPermutations > 0)
				{
					simTestThread.run("Permutations for AMOVA F-statics p-value");
				}

			double higherOrEqualValuesFsc = 0;
			double higherOrEqualValuesFst = 0;
			double higherOrEqualValuesFct = 0;
			
			for (int i = 0; i < numberOfPermutations; i++)
				{
					ArrayList<Node> permutationsNodes = new ArrayList<Node>();
					Permutations permutations = new Permutations();
					permutationsNodes.addAll(permutations.permutationsAMOVAFsc(nodes, groups));
					double[] Fstats1 = new double[9];
					Fstats1 = this.amovaFstatistics(locus, groups, permutationsNodes, true, false);
					double permutationsFsc = Fstats1[0];
					simTestThread.run("   Locus " + (h+1) + ", Fsc, permutation " + (i+1) + "       " + (((double)(Math.round(permutationsFsc*1000)))/1000));
					if (permutationsFsc < Fstats0[0])	{	}
					else
						{
							higherOrEqualValuesFsc = higherOrEqualValuesFsc + 1; // System.out.println("Significant Fsc = " + permutationsFsc);
						}
				}
			if (numberOfPermutations > 0)
				{
					pvalFsc = higherOrEqualValuesFsc/numberOfPermutations;
				}
			for (int i = 0; i < numberOfPermutations; i++)
				{
					ArrayList<Node> permutationsNodes = new ArrayList<Node>();
					Permutations permutations = new Permutations();
					permutationsNodes.addAll(permutations.permutationsGlabalGstNstFst(nodes));
					double[] Fstats2 = new double[9];
					Fstats2 = this.amovaFstatistics(locus, groups, permutationsNodes, true, false);
					double permutationsFst = Fstats2[1];
					simTestThread.run("   Locus " + (h+1) + ", Fst, permutation " + (i+1) + "       " + (float) (((double)(Math.round(permutationsFst*1000)))/1000));
					if (permutationsFst < Fstats0[1])	{	}
					else
						{
							higherOrEqualValuesFst = higherOrEqualValuesFst + 1;
						}
				}
			if (numberOfPermutations > 0)
				{
					pvalFst = higherOrEqualValuesFst/numberOfPermutations;
				}
			for (int i = 0; i < numberOfPermutations; i++)
				{
					ArrayList<Group> permutationsGroups = new ArrayList<Group>();
					Permutations permutations = new Permutations();
					permutationsGroups.addAll(permutations.permutationsAMOVAFct(groups));
					double[] Fstats3 = new double[9];
					Fstats3 = this.amovaFstatistics(locus, permutationsGroups, nodes, false, true);
					double permutationsFct = Fstats3[2];
					simTestThread.run("   Locus " + (h+1) + ", Fct, permutation " + (i+1) + "       " + (((double)(Math.round(permutationsFct*1000)))/1000));
					if (permutationsFct < Fstats0[2])	{	}
					else
						{
							higherOrEqualValuesFct = higherOrEqualValuesFct + 1;
						}
				}
			if (numberOfPermutations > 0)
				{
					pvalFct = higherOrEqualValuesFct/numberOfPermutations;
				}
				
			Fstats[0] = Fstats0[0]; Fstats[3] = pvalFsc; 
			Fstats[1] = Fstats0[1]; Fstats[4] = pvalFst;
			Fstats[2] = Fstats0[2]; Fstats[5] = pvalFct;
			
			Fstats[6] = Fstats0[3]; Fstats[7] = Fstats0[4]; 
			Fstats[8] = Fstats0[5]; Fstats[9] = Fstats0[6];
			Fstats[10] = Fstats0[7]; Fstats[11] = Fstats0[8];
			return Fstats;
		}
	
	// Method to compute AMOVA F-statistics (Excoffier et al, 1992):
	public double[] amovaFstatistics(Locus locus, ArrayList<Group> groups0, ArrayList<Node> nodes0, boolean permutPopID, boolean permutPopArrayListInGroups)
		{				
			double Fsc = 0; double Fst = 0; double Fct = 0;
			
			ArrayList<Group> groups = new ArrayList<Group>();
			ArrayList<Node> nodes = new ArrayList<Node>();
			double numberOfNodes = 0; double numberOfPopulations = 0;
			for (int g = 0; g < groups0.size(); g++)
				{
					Group group = new Group(groups0.get(g).groupID,groups0.get(g).name);
					for (int i = 0; i < groups0.get(g).populations.size(); i++)
						{
							if (permutPopID == true)
								{
									Population population = new Population(groups0.get(g).populations.get(i).popName, groups0.get(g).populations.get(i).popID);		
									for (int j = 0; j < nodes0.size(); j++)
										{
											if (population.popID == nodes0.get(j).permutPopID)
												{
													population.nodes.add(nodes0.get(j));
													nodes.add(nodes0.get(j));
												}
										}
									group.populations.add(population);
									numberOfNodes = numberOfNodes + population.nodes.size();
									numberOfPopulations = numberOfPopulations + 1;
								}
							else if (permutPopArrayListInGroups == true)
								{
									Population population = new Population(groups0.get(g).permutPopulations.get(i).popName, groups0.get(g).permutPopulations.get(i).popID);		
									for (int j = 0; j < nodes0.size(); j++)
										{
											if (population.popID == nodes0.get(j).currentPop)
												{
													population.nodes.add(nodes0.get(j));
													nodes.add(nodes0.get(j));
												}
										}
									group.populations.add(population);
									numberOfNodes = numberOfNodes + population.nodes.size();
									numberOfPopulations = numberOfPopulations + 1;
								}
							else
								{
									Population population = new Population(groups0.get(g).populations.get(i).popName, groups0.get(g).populations.get(i).popID);	
									for (int j = 0; j < nodes0.size(); j++)
										{
											if (population.popID == nodes0.get(j).currentPop)
												{
													population.nodes.add(nodes0.get(j));
													nodes.add(nodes0.get(j));
												}
										}
									group.populations.add(population);
									numberOfNodes = numberOfNodes + population.nodes.size();
									numberOfPopulations = numberOfPopulations + 1;
								}
						}
					groups.add(group);
				}
			double N = numberOfNodes; double P = numberOfPopulations; double G = (double) groups0.size();
			
			double SSDwp = 0; double ssdwp = 0;
			for (int g = 0; g < groups.size(); g++)
				{
					for (int i = 0; i < groups.get(g).populations.size(); i++)
						{
							ssdwp = 0;
							for (int j = 0; j < groups.get(g).populations.get(i).nodes.size(); j++)
								{
									for (int k = 0; k < groups.get(g).populations.get(i).nodes.size(); k++)
										{
											ssdwp = ssdwp + locus.nodesMismatches[groups.get(g).populations.get(i).nodes.get(j).nodeMismatchesIndex][groups.get(g).populations.get(i).nodes.get(k).nodeMismatchesIndex];
										}
								}
							SSDwp = SSDwp + ssdwp/(2*((double) groups.get(g).populations.get(i).nodes.size()));
						}		
				}
			
			double SSDapwg = 0;
			for (int g = 0; g < groups.size(); g++)
				{
					double SSDapwg1 = 0; double SSDapwg2 = 0;
					double Nk2 = 0;
					for (int i = 0; i < groups.get(g).populations.size(); i++)
						{
							for (int j = 0; j < groups.get(g).populations.get(i).nodes.size(); j++)
								{
									for (int k = 0; k < groups.get(g).populations.size(); k++)
										{
											for (int l = 0; l < groups.get(g).populations.get(k).nodes.size(); l++)
												{
													SSDapwg1 = SSDapwg1 + locus.nodesMismatches[groups.get(g).populations.get(i).nodes.get(j).nodeMismatchesIndex][groups.get(g).populations.get(k).nodes.get(l).nodeMismatchesIndex];
												}
										}
								}
							Nk2 = Nk2 + (2*((double) groups.get(g).populations.get(i).nodes.size()));
							double a = 0;
							for (int j = 0; j < groups.get(g).populations.get(i).nodes.size(); j++)
								{
									for (int k = 0; k < groups.get(g).populations.get(i).nodes.size(); k++)
										{
											a = a + locus.nodesMismatches[groups.get(g).populations.get(i).nodes.get(j).nodeMismatchesIndex][groups.get(g).populations.get(i).nodes.get(k).nodeMismatchesIndex];
										}
								}
							SSDapwg2 = SSDapwg2 + (a/(2*((double) groups.get(g).populations.get(i).nodes.size())));
						}
					SSDapwg1 = SSDapwg1/Nk2;
					SSDapwg = SSDapwg + (SSDapwg1 - SSDapwg2);	
				}
			
			double SSDag = 0;
			double SSDag1 = 0; double SSDag2 = 0;
			for (int i = 0; i < nodes.size(); i++)
				{
					for (int j = 0; j < nodes.size(); j++)
						{
							SSDag1 = SSDag1 + locus.nodesMismatches[nodes.get(i).nodeMismatchesIndex][nodes.get(j).nodeMismatchesIndex];
						}
				}
			SSDag1 = SSDag1/(2*N);
			for (int g = 0; g < groups.size(); g++)
				{
					double ssdag2 = 0; double Nk2 = 0;
					for (int i = 0; i < groups.get(g).populations.size(); i++)
						{
							for (int j = 0; j < groups.get(g).populations.get(i).nodes.size(); j++)
								{
									for (int k = 0; k < groups.get(g).populations.size(); k++)
										{
											for (int l = 0; l < groups.get(g).populations.get(k).nodes.size(); l++)
												{
													ssdag2 = ssdag2 + locus.nodesMismatches[groups.get(g).populations.get(i).nodes.get(j).nodeMismatchesIndex][groups.get(g).populations.get(k).nodes.get(l).nodeMismatchesIndex];
												}
										}
								}
							Nk2 = Nk2 + (2*((double) groups.get(g).populations.get(i).nodes.size()));
						}
					SSDag2 = SSDag2 + (ssdag2/Nk2);
				}
			SSDag = SSDag1 - SSDag2;
			
			// System.out.println("SSDwp = " + SSDwp);
			// System.out.println("SSDapwg = " + SSDapwg);
			// System.out.println("SSDag = " + SSDag);
			
			double MSDwp = SSDwp/(N-P);																					
			double MSDapwg = SSDapwg/(P-G);
			double MSDag = SSDag/(G-1);
			
			double Sg1 = 0; double Sg2 = 0; double Sg3 = 0;
			for (int g = 0; g < groups.size(); g++)
				{
					double Ng = 0;
					for (int i = 0; i < groups.get(g).populations.size(); i++) 
						{
							Ng = Ng + groups.get(g).populations.get(i).nodes.size();
						}
					
					for (int i = 0; i < groups.get(g).populations.size(); i++)
						{
							double Nk = groups.get(g).populations.get(i).nodes.size();
							Sg1 = Sg1 + ((Nk*Nk)/Ng);
						}
					for (int i = 0; i < groups.get(g).populations.size(); i++) 
						{
							double Nk = groups.get(g).populations.get(i).nodes.size();
							Sg2 = Sg2 + ((Nk*Nk)/N);
						}
					Sg3 = Sg3 + ((Ng*Ng)/N);
				}
			
			double n = (N-Sg1)/(P-G);
			double nPrime = (Sg1-Sg2)/(G-1);
			double nPrimePrime = (N-Sg3)/(G-1); 
					
			double Oc = MSDwp;
			double Ob = (MSDapwg-Oc)/n;
			double Oa = (MSDag-(nPrime*Ob)- Oc)/nPrimePrime;
			double O = Oa + Ob + Oc;
			
			Fsc = Ob/(Ob+Oc); // System.out.println("Fsc = " + Fsc);
			Fst = (Oa+Ob)/O; // System.out.println("Fst = " + Fst);
			Fct = Oa/O; // System.out.println("Fct = " + Fct);
			
			double[] Fstats = new double[9];
			Fstats[0] = Fsc;
			Fstats[1] = Fst;
			Fstats[2] = Fct;
			
			Fstats[3] = Ob; Fstats[4] = Ob+Oc;
			Fstats[5] = Oa+Ob; Fstats[6] = O;
			Fstats[7] = Oa; Fstats[8] = O;
			return Fstats;
		}
	
	// Method to (ONLY) compute the AMOVA Fct (Excoffier et al, 1992):
	/* public double amovaFct(Locus locus, ArrayList<Group> groups, ArrayList<Node> nodes)
		{				
			double Fct = 0;
			double numberOfPopulations = 0;
			for (int g = 0; g < groups.size(); g++)
				{
					numberOfPopulations = numberOfPopulations + groups.get(g).populations.size();
				}
			double N = (double) nodes.size(); double P = numberOfPopulations; double G = (double) groups.size();
			
			double SSDwp = 0; double ssdwp = 0;
			for (int g = 0; g < groups.size(); g++)
				{
					for (int i = 0; i < groups.get(g).populations.size(); i++)
						{
							ssdwp = 0;
							for (int j = 0; j < groups.get(g).populations.get(i).nodes.size(); j++)
								{
									for (int k = 0; k < groups.get(g).populations.get(i).nodes.size(); k++)
										{
											ssdwp = ssdwp + locus.nodesMismatches[groups.get(g).populations.get(i).nodes.get(j).nodeMismatchesIndex][groups.get(g).populations.get(i).nodes.get(k).nodeMismatchesIndex];
										}
								}
							SSDwp = SSDwp + ssdwp/(2*((double) groups.get(g).populations.get(i).nodes.size()));
						}		
				}
			
			double SSDapwg = 0;
			for (int g = 0; g < groups.size(); g++)
				{
					double SSDapwg1 = 0; double SSDapwg2 = 0;
					double Nk2 = 0;
					for (int i = 0; i < groups.get(g).populations.size(); i++)
						{
							for (int j = 0; j < groups.get(g).populations.get(i).nodes.size(); j++)
								{
									for (int k = 0; k < groups.get(g).populations.size(); k++)
										{
											for (int l = 0; l < groups.get(g).populations.get(k).nodes.size(); l++)
												{
													SSDapwg1 = SSDapwg1 + locus.nodesMismatches[groups.get(g).populations.get(i).nodes.get(j).nodeMismatchesIndex][groups.get(g).populations.get(k).nodes.get(l).nodeMismatchesIndex];
												}
										}
								}
							Nk2 = Nk2 + (2*((double) groups.get(g).populations.get(i).nodes.size()));
							double a = 0;
							for (int j = 0; j < groups.get(g).populations.get(i).nodes.size(); j++)
								{
									for (int k = 0; k < groups.get(g).populations.get(i).nodes.size(); k++)
										{
											a = a + locus.nodesMismatches[groups.get(g).populations.get(i).nodes.get(j).nodeMismatchesIndex][groups.get(g).populations.get(i).nodes.get(k).nodeMismatchesIndex];
										}
								}
							SSDapwg2 = SSDapwg2 + (a/(2*((double) groups.get(g).populations.get(i).nodes.size())));
						}
					SSDapwg1 = SSDapwg1/Nk2;
					SSDapwg = SSDapwg + (SSDapwg1 - SSDapwg2);	
				}
			
			double SSDag = 0;
			double SSDag1 = 0; double SSDag2 = 0;
			for (int i = 0; i < nodes.size(); i++)
				{
					for (int j = 0; j < nodes.size(); j++)
						{
							SSDag1 = SSDag1 + locus.nodesMismatches[nodes.get(i).nodeMismatchesIndex][nodes.get(j).nodeMismatchesIndex];
						}
				}
			SSDag1 = SSDag1/(2*N);
			for (int g = 0; g < groups.size(); g++)
				{
					double ssdag2 = 0; double Nk2 = 0;
					for (int i = 0; i < groups.get(g).populations.size(); i++)
						{
							for (int j = 0; j < groups.get(g).populations.get(i).nodes.size(); j++)
								{
									for (int k = 0; k < groups.get(g).populations.size(); k++)
										{
											for (int l = 0; l < groups.get(g).populations.get(k).nodes.size(); l++)
												{
													ssdag2 = ssdag2 + locus.nodesMismatches[groups.get(g).populations.get(i).nodes.get(j).nodeMismatchesIndex][groups.get(g).populations.get(k).nodes.get(l).nodeMismatchesIndex];
												}
										}
								}
							Nk2 = Nk2 + (2*((double) groups.get(g).populations.get(i).nodes.size()));
						}
					SSDag2 = SSDag2 + (ssdag2/Nk2);
				}
			SSDag = SSDag1 - SSDag2;
			
			// System.out.println("SSDwp = " + SSDwp);
			// System.out.println("SSDapwg = " + SSDapwg);
			// System.out.println("SSDag = " + SSDag);
			
			double MSDwp = SSDwp/(N-P);																					
			double MSDapwg = SSDapwg/(P-G);
			double MSDag = SSDag/(G-1);
			
			double Sg1 = 0; double Sg2 = 0; double Sg3 = 0;
			for (int g = 0; g < groups.size(); g++)
				{
					double Ng = 0;
					for (int i = 0; i < groups.get(g).populations.size(); i++) 
						{
							Ng = Ng + groups.get(g).populations.get(i).nodes.size();
						}
					
					for (int i = 0; i < groups.get(g).populations.size(); i++)
						{
							double Nk = groups.get(g).populations.get(i).nodes.size();
							Sg1 = Sg1 + ((Nk*Nk)/Ng);
						}
					for (int i = 0; i < groups.get(g).populations.size(); i++) 
						{
							double Nk = groups.get(g).populations.get(i).nodes.size();
							Sg2 = Sg2 + ((Nk*Nk)/N);
						}
					Sg3 = Sg3 + ((Ng*Ng)/N);
				}
			
			double n = (N-Sg1)/(P-G);
			double nPrime = (Sg1-Sg2)/(G-1);
			double nPrimePrime = (N-Sg3)/(G-1); 
					
			double Oc = MSDwp;
			double Ob = (MSDapwg-Oc)/n;
			double Oa = (MSDag-(nPrime*Ob)- Oc)/nPrimePrime;
			double O = Oa + Ob + Oc;
			
			Fct = Oa/O; // System.out.println("Fct = " + Fct);
			return Fct;
		} */
}
