package spads1;

// Class created by Simon Dellicour (June 2011)

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.triangulate.DelaunayTriangulationBuilder;

public class InputFileReader 
{
	private Interface interface1;
	public double permutationsGstNst;
	public double permutationsFst;
	public double permutationsAMOVA;
	public double minimumNumberOfSamovaGroups;
	public double maximumNumberOfSamovaGroups;
	public double numberOfSamovaIterations;
	public double numberOfSamovaRepetitions;
	public double minimumNumberOfSamovamGroups;
	public double maximumNumberOfSamovamGroups;
	public double numberOfSamovamIterations;
	public double numberOfSamovamRepetitions;
	public double minimumNumberOfBarriers;
	public double maximumNumberOfBarriers;
	public double minimumNumberOfBarriersm;
	public double maximumNumberOfBarriersm;
	public boolean sameNodesInDifferentGroups;
	public boolean pairwisePopulationDistances;
	public boolean pairwiseMonmonierDistances;
	public double[][] populationDistances;
	public double[][] monmonierDistances;
	public ArrayList<String> populationNames;
	public double numberOfLoci;
	public double numberOfNodes;
	public boolean ploidy;
	public String inputName;
	public ArrayList<Locus> loci;
	public ArrayList<String> externalMessage;
	public SimThread simThread;
	
	public InputFileReader(Interface window, SimThread simThread0) throws IOException, InterruptedException 
		{
			interface1 = new Interface(); interface1.equals(window);
			double numberOfUserGroups = window.getNumberOfUserGroups();
			permutationsGstNst = window.getPermutationsGstNst();
			permutationsFst = window.getPermutationsFst();
			permutationsAMOVA = window.getPermutationsAMOVA();
			minimumNumberOfSamovaGroups = window.getMinimumNumberOfSamovaGroups();
			maximumNumberOfSamovaGroups = window.getMaximumNumberOfSamovaGroups();
			numberOfSamovaIterations = window.getSamovaNumberOfIterations();
			numberOfSamovaRepetitions = window.getSamovaNumberOfRepetitions();
			numberOfSamovaIterations = window.getSamovaNumberOfIterations();
			numberOfSamovaRepetitions = window.getSamovaNumberOfRepetitions();
			minimumNumberOfBarriers = window.getMinimumNumberOfBarriers();
			maximumNumberOfBarriers = window.getMaximumNumberOfBarriers();
			sameNodesInDifferentGroups = false;
			numberOfLoci = window.getNumberOfLoci();
			numberOfNodes = 0;
			inputName = window.getInputName(); 
			loci = new ArrayList<Locus>();	
			externalMessage = new ArrayList<String>();
			simThread = simThread0;
			
			if (inputName.equals("summary_stats_comparison"))
				{	}
			else
				{
					externalMessage.add("Number of user-defined groups: " + (int) numberOfUserGroups + ".");
					externalMessage.add("Number of loci: " + (int) numberOfLoci + ".");
					externalMessage.add("Minimum number of SAMOVA groups to determine: " + (int) minimumNumberOfSamovaGroups + ".");
					externalMessage.add("Maximum number of SAMOVA groups to determine: " + (int) maximumNumberOfSamovaGroups + ".");
					externalMessage.add("Minimum number of barriers to construct (with the Monmonier algorithm): " + (int) minimumNumberOfBarriers + ".");
					externalMessage.add("Maximum number of barriers to construct (with the Monmonier algorithm): " + (int) maximumNumberOfBarriers + ".");
					externalMessage.add("	");
					for (int i = 0; i < (int) numberOfLoci; i++)
						{
							Locus locus = new Locus(i+1);
							ArrayList<String> lines1 = new ArrayList<String>();
							ArrayList<String> lines2 = new ArrayList<String>();
							ArrayList<String> lines3 = new ArrayList<String>();		
							
							try {
									String name1 = inputName + "_populations.txt";
									BufferedReader in1 = new BufferedReader(new FileReader(name1));
									String s1;
									ArrayList<String> lines10 = new ArrayList<String>();
									while((s1 = in1.readLine()) != null) lines10.add(s1);
									lines1 = lines10;
									in1.close();
								}
							catch (Exception e1)
								{
									simThread.run("SPADS does not find the populations file");
									externalMessage.add("SPADS does not find the populations file");
								}
							
							for (int j = 0; j < lines1.size(); j ++)
								{
									if (lines1.get(j).isEmpty())
										{ }
									else
										{
											String line0 = (String) lines1.get(j);
											String[] line = line0.split("[ 	]+");
											String name = new String(); name = (String) line[0];
											int x = j+1;
											Population population = new Population(name, x);
											population.position[0] = (double) Double.parseDouble((String) line[1]);
											population.position[1] = (double) Double.parseDouble((String) line[2]);
											population.popDistanceIndex = j;
											locus.populations.add(population);
										}
								}
							
							String name2 = inputName + "_locus" + (i+1) + ".phy";
							try {
									BufferedReader in2 = new BufferedReader(new FileReader(name2));
									String s2;
									ArrayList<String> lines20 = new ArrayList<String>();
									while((s2 = in2.readLine()) != null) lines20.add(s2);
									lines2 = lines20;
									in2.close();
								}
							catch (Exception e1)
								{
									simThread.run("SPADS does not find the Phylip file for locus " + (i+1));
									externalMessage.add("SPADS does not find the Phylip file for locus " + (i+1));
								}
							
							String line01 = (String) lines2.get(0);
							String[] line1 = line01.split("[ 	]+");
							locus.length = Double.parseDouble((String) line1[1]);
							ploidy = false;
							if (line1.length > 2) 
								{	
									locus.ploidy = Double.parseDouble((String) line1[2]);
									ploidy = true;
								}							
							
							numberOfNodes = 0;
							for (int j = 1; j < lines2.size(); j++)
								{
									if (lines2.get(j).isEmpty())
										{ }
									else
										{
											numberOfNodes = numberOfNodes + 1;
											String line0 = (String) lines2.get(j);
											String[] line = line0.split(" ");
											
											String id0 = new String(); id0 = line[0];
											String[] id = id0.split("-");
											String nodeName = new String(); nodeName = (String) id[0];
											String popName = new String(); popName = (String) id[1];
											int x = j; 
											int y = 0;
											for (int k = 0; k < locus.populations.size(); k++)
												{
													if (locus.populations.get(k).popName.equals(popName))
														{
															y = locus.populations.get(k).popID;
														}
												}
											Node node = new Node(nodeName, x, popName, y);				
											int k = 1;
											while (line[k].equalsIgnoreCase(""))
												{
													k = k + 1;
												}
											node.finalSequence = line[k];
											String[] sequence0 = new String[node.finalSequence.split("").length];
											node.sequence = new ArrayList<String>();
											sequence0 =	node.finalSequence.split("");
											for (int l1 = 0; l1 < sequence0.length; l1++)
												{
													if (sequence0[l1].isEmpty() | sequence0[l1] == " ")	
														{	}
													else
														{
															node.sequence.add(sequence0[l1]);
														}
												}
											for (int l = 0; l < locus.populations.size(); l++)
												{
													if (node.currentPop == locus.populations.get(l).popID)
														{
															locus.populations.get(l).nodes.add(node);
														}
												}							
										}
								}
							
							if (ploidy == true)
								{
									int indID = 0;
									for (int j = 0; j < locus.populations.size(); j++)
										{
											ArrayList<Individual> individuals = new ArrayList<Individual>();
											for (int k = 0; k < locus.populations.get(j).nodes.size(); k++)
												{
													String fullName1 = locus.populations.get(j).nodes.get(k).nodeName;
													String[] fullName2 = new String[fullName1.split(",").length];
													fullName2 = fullName1.split("\\.");
													String name = (String) fullName2[0];
													boolean newIndividual = true;
													boolean sequenceInsertion = false;
													for (int l = 0; l < individuals.size(); l++)
														{
															if (name.equals(individuals.get(l).indName))
																{
																	newIndividual = false;
																	if (sequenceInsertion == false)
																		{
																			sequenceInsertion = true;
																			individuals.get(l).nodes.add(locus.populations.get(j).nodes.get(k));
																		}
																}
														}
													if (newIndividual == true)
														{
															indID = indID + 1;
															Individual individual = new Individual(name,indID);
															individual.nodes.add(locus.populations.get(j).nodes.get(k));
															individual.locusID = locus.ID;
															individuals.add(individual);
														}
												}
											locus.populations.get(j).individuals = new ArrayList<Individual>();
											locus.populations.get(j).individuals.addAll(individuals);
										}
									indID = 0;
									for (int j = 0; j < locus.populations.size(); j++)
										{
											ArrayList<Individual> individualsGhosts = new ArrayList<Individual>();
											for (int k = 0; k < locus.populations.get(j).nodes.size(); k++)
												{
													String fullName1 = locus.populations.get(j).nodes.get(k).nodeName;
													String[] fullName2 = new String[fullName1.split(",").length];
													fullName2 = fullName1.split("\\.");
													String name = (String) fullName2[0];
													boolean newIndividual = true;
													boolean sequenceInsertion = false;
													for (int l = 0; l < individualsGhosts.size(); l++)
														{
															if (name.equals(individualsGhosts.get(l).indName))
																{
																	newIndividual = false;
																	if (sequenceInsertion == false)
																		{
																			sequenceInsertion = true;
																			individualsGhosts.get(l).nodes.add(locus.populations.get(j).nodes.get(k));
																		}
																}
														}
													if (newIndividual == true)
														{
															indID = indID + 1;
															Individual individual = new Individual(name,indID);
															individual.nodes.add(locus.populations.get(j).nodes.get(k));
															individualsGhosts.add(individual);
														}
												}
											locus.populations.get(j).individualsGhosts = new ArrayList<Individual>();
											locus.populations.get(j).individualsGhosts.addAll(individualsGhosts);
										}
								}
							
							// To remove the populations without any Node objects:
							// (this could arrive when several loci are linked to the same "input_populations.txt" file
							ArrayList<Population> trueLocusPopulations = new ArrayList<Population>();
							for (int l = 0; l < locus.populations.size(); l++)
								{
									if (locus.populations.get(l).nodes.size() == 0)
										{	}
									else
										{
											trueLocusPopulations.add(locus.populations.get(l));
										}
								}
							locus.populations = new ArrayList<Population>();
							locus.populations.addAll(trueLocusPopulations);
							
							if (numberOfUserGroups > 0)
								{
									try {
											String name3 = inputName + "_groups.txt";
											BufferedReader in3 = new BufferedReader(new FileReader(name3));
											String s3;
											ArrayList<String> lines30 = new ArrayList<String>();
											while((s3 = in3.readLine()) != null) lines30.add(s3);
											lines3 = lines30;
											in3.close();
										}
									catch (Exception e1)
										{
											simThread.run("SPADS does not find the groups file");
											externalMessage.add("SPADS does not find the groups file");
										}
										
									for (int j = 0; j < lines3.size(); j++)
										{
											if (lines3.get(j).isEmpty())
												{ }
											else
												{
													String line0 = (String) lines3.get(j);
													String[] line = line0.split("[ 	]+");
													String name = new String(); name = (String) line[0];
													int x = j+1;
													Group group = new Group(x, name);
													
													for (int k = 0; k < locus.populations.size(); k++)
														{
															for (int l = 1; l < line.length; l++)
																{
																	String popName = (String) line[l];
																	if (locus.populations.get(k).popName.equals(popName))
																		{
																			group.populations.add(locus.populations.get(k));
																			group.nodes.addAll(locus.populations.get(k).nodes);
																		}
																}
														}
													// To remove the groups without any Population objects:
													// (this could arrive when several loci are linked to the same "input_groups.txt" file
													if (group.populations.size() > 0)
														{	
															locus.userGroups.add(group);
														}
												}
										}
									for (int j = 1; j < locus.userGroups.size(); j++)
										{
											for (int k = 0; k < j; k++)
												{
													for (int l = 0; l < locus.userGroups.get(j).populations.size(); l++)
														{
															for (int m = 0; m < locus.userGroups.get(k).populations.size(); m++)
																{	
																	if (locus.userGroups.get(j).populations.get(l).popName.equals(locus.userGroups.get(k).populations.get(m).popName))
																		{
																			sameNodesInDifferentGroups = true;
																		}
																}
														}
												}
										}
										
								}
							
							if (minimumNumberOfSamovaGroups > 0 || minimumNumberOfSamovamGroups > 0 || minimumNumberOfBarriers > 0 || minimumNumberOfBarriersm > 0)
								{
									locus.delaunayTriangulation = new int[(int) locus.populations.size()][(int) locus.populations.size()];
									for (int j = 0; j < locus.populations.size(); j++)
										{
											for (int k = 0; k < locus.populations.size(); k++)
												{
													locus.delaunayTriangulation[j][k] = 0;
												}
										}
									locus.delaunayTriangulation = this.delaunayTriangulationConstruction(locus, locus.populations);
									
									if (locus.ID == 1)
										{
											externalMessage.add("Delaunay triangulation used for the 1st locus (1 corresponds to a triangulation connection between two sampled populations): ");
											externalMessage.add("	");
											String popID1 = new String();
											ArrayList<String> popID2 = new ArrayList<String>();
											for (int j = 0; j < locus.populations.size(); j++)
												{
													popID1 = popID1 + locus.populations.get(j).popName + "	";
													popID2.add(locus.populations.get(j).popName);
												}
											externalMessage.add("	" + popID1);
											for (int j = 0; j < locus.populations.size(); j++)
												{	
													String line = new String();
													line = (popID2.get(j) + "	");
													for (int k = 0; k < locus.populations.size(); k++)
														{
															line = line + (locus.delaunayTriangulation[j][k] + "	");
														}	externalMessage.add(line);
												}	externalMessage.add("	");
										}
								}
						
							loci.add(locus);
						}
					
					ArrayList<String> lines1 = new ArrayList<String>();
					String name1 = inputName + "_populations.txt";
					BufferedReader in1 = new BufferedReader(new FileReader(name1));
					String s1;
					ArrayList<String> lines10 = new ArrayList<String>();
					while((s1 = in1.readLine()) != null) lines10.add(s1);
					lines1 = lines10;
					in1.close();
					populationNames = new ArrayList<String>(); String name = new String();
					for (int i = 0; i < lines1.size(); i++)
						{
							String line1 = (String) lines1.get(i);
							String[] line = line1.split("[ 	]+");
							name = new String(); name = (String) line[0];
							populationNames.add(name);
						}
					try {
							ArrayList<String> lines4 = new ArrayList<String>();
							String name4 = inputName + "_population_distances.txt";
							BufferedReader in4 = new BufferedReader(new FileReader(name4));
							String s4;
							ArrayList<String> lines40= new ArrayList<String>();
							while((s4 = in4.readLine()) != null) lines40.add(s4);
							lines4 = lines40;
							in4.close();
							pairwisePopulationDistances = true;
							boolean matrixWithGoodDimensions = true;
							populationDistances = new double[lines1.size()][lines1.size()];

							if (lines4.size() == lines1.size())
								{ }
							else
								{
									matrixWithGoodDimensions = false;
								}
							for (int i = 0; i < lines4.size(); i++)
								{
									if (lines4.get(i).isEmpty())
										{ }
									else
										{
											String line0 = (String) lines4.get(i);
											String[] line = line0.split("[ 	]+");
											if (line.length == lines1.size())
												{
													for (int j = 0; j < line.length; j++)
														{
															populationDistances[i][j] = Double.parseDouble(line[j]);
														}
												}
											else
												{
													matrixWithGoodDimensions = false;
												}
										}
								}
							if (matrixWithGoodDimensions == false)
								{
									pairwisePopulationDistances = false;
									simThread.run("   Please verify its dimensions.");
									simThread.run("   pairwise geographic distances.");
									simThread.run("There is a problem with the matrix of population");
									simThread.stopThread();
								}
							}
						catch (Exception e1)
							{
								pairwisePopulationDistances = false;
							}	
						try {
								ArrayList<String> lines4 = new ArrayList<String>();
								String name4 = inputName + "_monmonier_distances.txt";
								BufferedReader in4 = new BufferedReader(new FileReader(name4));
								String s4;
								ArrayList<String> lines40= new ArrayList<String>();
								while((s4 = in4.readLine()) != null) lines40.add(s4);
								lines4 = lines40;
								in4.close();
								pairwiseMonmonierDistances = true;
								boolean matrixWithGoodDimensions = true;
								monmonierDistances = new double[lines1.size()][lines1.size()];
		
								if (lines4.size() == lines1.size())
									{ }
								else
									{
										matrixWithGoodDimensions = false;
									}
								for (int i = 0; i < lines4.size(); i++)
									{
										if (lines4.get(i).isEmpty())
											{ }
										else
											{
												String line0 = (String) lines4.get(i);
												String[] line = line0.split("[ 	]+");
												if (line.length == lines1.size())
													{
														for (int j = 0; j < line.length; j++)
															{
																monmonierDistances[i][j] = Double.parseDouble(line[j]);
															}
													}
												else
													{
														matrixWithGoodDimensions = false;
													}
											}
									}
								if (matrixWithGoodDimensions == false)
									{
										pairwiseMonmonierDistances = false;
										simThread.run("   Please verify its dimensions.");
										simThread.run("   pairwise distances for Monmonier algorithm.");
										simThread.run("There is a problem with the matrix of population");
										simThread.stopThread();
									}
							}
						catch (Exception e1)
							{
								pairwisePopulationDistances = false;
							}
						if ((sameNodesInDifferentGroups == true) & window.checkBoxAMOVA.isSelected())
							{
								simThread.run("   AMOVA analysis has been cancelled.");
								simThread.run("   in more than one group.");
								simThread.run("Warning: some populations are assigned");
								externalMessage.add("Warning: some populations are assigned in more than one group.");
								externalMessage.add("	AMOVA analysis has been cancelled");
							}
					}
		}
	
	// Method to compute the distance (dgeo) between 2 populations from a "populations" ArrayList:
	public double populationsDistance(ArrayList<Population> populations, int popID1, int popID2)
		// Important note: popID1 and popID2 are NOT some population ID BUT simply indices for the "populations" ArrayList !!!
		{
			double distance = -1;
			if (pairwisePopulationDistances == false)
				{
					double[] position1 = new double[2]; double[] position2 = new double[2];
					position1 = populations.get(popID1).position;
					position2 = populations.get(popID2).position;
					distance = Math.sqrt(((position2[1]-position1[1])*(position2[1]-position1[1])) + ((position2[0]-position1[0])*(position2[0]-position1[0])));
				}
			else
				{
					distance = populationDistances[populations.get(popID1).popDistanceIndex][populations.get(popID2).popDistanceIndex];
				}
			return distance; 
		}
	
	// Method to compute a triangulation of Delaunay between a given set of populations
	public int[][] delaunayTriangulationConstruction(Locus locus, ArrayList<Population> populations)
		{
			if (populations.size() == 1)
				{
					System.out.println("There is olny one sampled population");
					System.out.println("(no SAMOVA analysis).");
				}
			else if (populations.size() == 2)
				{
					locus.delaunayTriangulation[0][0] = 1;
				}
			else
				{
					DelaunayTriangulationBuilder delaunay = new DelaunayTriangulationBuilder();
					Collection<Coordinate> coordinates = new ArrayList<Coordinate>();
					Coordinate coor = new Coordinate(populations.get(0).position[0], populations.get(0).position[1]); coordinates.add(coor);
					for (int i = 1; i < populations.size(); i++)
						{
							coor = new Coordinate(populations.get(i).position[0], populations.get(i).position[1]); coordinates.add(coor);
						}
					delaunay.setSites(coordinates);
					GeometryFactory fact = new GeometryFactory(); 
					Geometry triangles = delaunay.getTriangles(fact);
					double[][] edges = new double[triangles.getNumGeometries()*3][4];
					for (int i = 0; i < triangles.getNumGeometries(); i++)
						{
							edges[(i*3)+0][0] = triangles.getGeometryN(i).getCoordinates()[1].x;
							edges[(i*3)+0][1] = triangles.getGeometryN(i).getCoordinates()[1].y;
							edges[(i*3)+0][2] = triangles.getGeometryN(i).getCoordinates()[2].x;
							edges[(i*3)+0][3] = triangles.getGeometryN(i).getCoordinates()[2].y;
							edges[(i*3)+1][0] = triangles.getGeometryN(i).getCoordinates()[2].x;
							edges[(i*3)+1][1] = triangles.getGeometryN(i).getCoordinates()[2].y;
							edges[(i*3)+1][2] = triangles.getGeometryN(i).getCoordinates()[3].x;
							edges[(i*3)+1][3] = triangles.getGeometryN(i).getCoordinates()[3].y;
							edges[(i*3)+2][0] = triangles.getGeometryN(i).getCoordinates()[3].x;
							edges[(i*3)+2][1] = triangles.getGeometryN(i).getCoordinates()[3].y;
							edges[(i*3)+2][2] = triangles.getGeometryN(i).getCoordinates()[1].x;
							edges[(i*3)+2][3] = triangles.getGeometryN(i).getCoordinates()[1].y;	
						}
					for (int i = 0; i < (triangles.getNumGeometries()*3); i++)
						{	
							for (int j = 0; j < populations.size(); j++)
								{
									// System.out.println(populations.get(j).position[0] + "	" + edges[i][0] + "	" + populations.get(j).position[1] + "	" + edges[i][1]);
									if ((populations.get(j).position[0] == edges[i][0]) & (populations.get(j).position[1] == edges[i][1]))
										{
											// System.out.println("inside");
											for (int k = 0; k < populations.size(); k++)
												{
													if ((populations.get(k).position[0] == edges[i][2]) & (populations.get(k).position[1] == edges[i][3]))
														{
															locus.delaunayTriangulation[j][k] = 1;
															locus.delaunayTriangulation[k][j] = 1;
														}
												}
										}
								}
						}
				}
			return locus.delaunayTriangulation;
		}
}
