package spads1;

// Class created by Simon Dellicour (June 2011)

import java.util.ArrayList;
import javax.swing.JFrame;

public class Samova extends JFrame
{
	private static final long serialVersionUID = 1L;
	private InputFileReader input;
	private Amova amova;
	public ArrayList<String> results;
	public ArrayList<String> externalMessage;
	public SimThread simTestThread;
	
	// Population groups/clusters identification with the SAMOVA algorithm (Dupanloup et al, 2002):
	public Samova(InputFileReader input0, Interface window, SimThread simTestThread0) throws InterruptedException 
		{	
			simTestThread = simTestThread0;
			simTestThread.run("Beginning of locus by locus SAMOVA analysis");
			input = input0;
			amova = new Amova();
			results = new ArrayList<String>();
			externalMessage = new ArrayList<String>();
			int Kmin = 2;
			if (input.minimumNumberOfSamovaGroups > 1)
				{
					Kmin = (int) input.minimumNumberOfSamovaGroups;
				}
			else 
				{
					Kmin = 2;
				}
			double K = input.maximumNumberOfSamovaGroups;	

			results.add("SAMOVA RESULTS (locus by locus): ");
			results.add("");
			for (int h = 0; h < input.numberOfLoci; h++)
				{
					results.add("LOCUS " + input.loci.get(h).ID);
					ArrayList<Node> nodes = new ArrayList<Node>();
					nodes = input.loci.get(h).nodes;
					String popNamesString = new String();
					for (int k = 0; k < input.loci.get(h).populations.size(); k++)
						{
							popNamesString = popNamesString + input.loci.get(h).populations.get(k).popName + "	";
						}
					results.add("Clusters ID	" + popNamesString);
					
					for (int k = Kmin; k < (K+1); k++)
						{
							String finalGs = new String();
							double finalFct = -10;
							ArrayList<Double> allTheFct = new ArrayList<Double>();
							ArrayList<Group> finalGroups = new ArrayList<Group>();
							int[][] delaunayTriangulation = new int[input.loci.get(h).populations.size()][input.loci.get(h).populations.size()];
							for (int r = 0; r < input.numberOfSamovaRepetitions; r++)
								{
									Thread.sleep(10);
									simTestThread.run("	Locus " + input.loci.get(h).ID + ", K = " + k + ", repetition " + (r+1));
									double Fct = 0;	
									ArrayList<Group> groups = new ArrayList<Group>();
									delaunayTriangulation = new int[input.loci.get(h).populations.size()][input.loci.get(h).populations.size()];
									for (int i = 0; i < input.loci.get(h).populations.size(); i++)
										{
											for (int j = 0; j < input.loci.get(h).populations.size(); j++)
												{
													delaunayTriangulation[i][j] = input.loci.get(h).delaunayTriangulation[i][j]; // System.out.print(delaunayTriangulation[i][j] + " ");
												}	// System.out.println();
										}			// System.out.println();
									for (int i = 0; i < input.loci.get(h).populations.size(); i++)
										{
											input.loci.get(h).populations.get(i).samovaGroupID = -1;
										}
									
									// (9.1) Voronoi diagram construction: already done (during the input reading step)
											// Each Voronoi barrier corresponds to a Delaunay segment ("1" in the delaunayTriangulation matrix) between 2 sampled populations
									// (9.2 & 3) Arbitrary first partition of the sampled populations into K groups:
											// (here: each group, except one, is composed of a single population and the last group contains the populations not assigned to any other groups)
									for (int i = 0; i < k; i++)
										{
											double R3 = 0;
											int iInt = -1;
											for (int j = 0; j < input.loci.get(h).populations.size(); j++)
												{
													double r3 = 0; r3 = Math.random();
													if (input.loci.get(h).populations.get(j).samovaGroupID == -1
															&& r3 > R3)
														{
															R3 = r3;
															iInt = j;
														}
												}
											input.loci.get(h).populations.get(iInt).samovaGroupID = (i+1);
											Double ID = new Double(i+1); String groupName = new String(); groupName = ID.toString();
											Group group = new Group(i+1, groupName); group.populations.add(input.loci.get(h).populations.get(iInt));
											groups.add(group);
											for (int j = 0; j < input.loci.get(h).populations.size(); j++)
												{
													if (delaunayTriangulation[iInt][j] == 1)
														{
															delaunayTriangulation[iInt][j] = -1;
															delaunayTriangulation[j][iInt] = -1;	
														}
												}
										}
									double counter = 0; counter = k;
									while (counter != input.loci.get(h).populations.size())
										{
											for (int i = 0; i < input.loci.get(h).populations.size(); i++)
												{
													if (input.loci.get(h).populations.get(i).samovaGroupID == -1)
														{
															// System.out.println(input.loci.get(h).populations.get(i).popName);
															double R4 = 0; int jInt = -1;
															for (int j = 0; j < input.loci.get(h).populations.size(); j++)
																{
																	double r4 = 0;	r4 = Math.random();
																	if (delaunayTriangulation[i][j] == -1 
																			&& input.loci.get(h).populations.get(j).samovaGroupID > 0
																			&& r4 > R4)
																		{
																			R4 = r4;
																			jInt = j;		
																		}
																}
															if (R4 > 0)
																{
																	counter = counter + 1;
																	input.loci.get(h).populations.get(i).samovaGroupID = input.loci.get(h).populations.get(jInt).samovaGroupID;
																	delaunayTriangulation[i][jInt] = 1;
																	delaunayTriangulation[jInt][i] = 1;
																	for (int k1 = 0; k1 < input.loci.get(h).populations.size(); k1++)
																		{
																			if (delaunayTriangulation[i][k1] == 0) 
																				{	}
																			else if (delaunayTriangulation[i][k1] == 1)
																				{
																					if (input.loci.get(h).populations.get(i).samovaGroupID == input.loci.get(h).populations.get(k1).samovaGroupID)
																						{	}
																					else
																						{
																							delaunayTriangulation[i][k1] = -1;
																							delaunayTriangulation[k1][i] = -1;
																						}
																				}
																			else if (delaunayTriangulation[i][k1] == -1)
																				{
																					if (input.loci.get(h).populations.get(i).samovaGroupID != input.loci.get(h).populations.get(k1).samovaGroupID)
																						{	}
																					else
																						{
																							delaunayTriangulation[i][k1] = 1;
																							delaunayTriangulation[k1][i] = 1;
																						}
																				}
																		}
																	
																	for (int k1 = 0; k1 < groups.size(); k1++)
																		{
																			if (groups.get(k1).groupID == input.loci.get(h).populations.get(i).samovaGroupID)
																				{
																					groups.get(k1).populations.add(input.loci.get(h).populations.get(i));
																				}
																		}
																}
														}		
												}
										}
									/* for (int i = 0; i < input.loci.get(h).populations.size(); i++)
										{
											System.out.print(input.loci.get(h).populations.get(i).samovaGroupID);
										}	System.out.println(); */
									
									// (9.4) Computation of the Fct associated to these first K groups:
										// Fct = this.groupsFct(input.loci.get(h), groups, nodes); 
									double[] Fstats = new double[9];
									Fstats = amova.amovaFstatistics(input.loci.get(h), groups, nodes, false, false);
									Fct = Fstats[2];
									
									boolean newFctValue = false;
									// (9.5) Selection of a random edge: and the 2 sampled populations located on both sides of this selected edge are identified,
											// and one population chosen at random is assigned to the group of the other population:
									for (int numberOfSteps = 0; numberOfSteps < input.numberOfSamovaIterations; numberOfSteps++) // IT SHOULD BE 10'000 steps like in the SAMOVA algorithm !!!!!!!!!!!!!!!
										{
											double newFct = 0;
											int[] randomEdge = new int[2];
											double R = 1;
											for (int i = 0; i < input.loci.get(h).populations.size(); i++)
												{
													for (int j = 0; j < input.loci.get(h).populations.size(); j++)
														{
															if (delaunayTriangulation[i][j] == -1 &&
																	groups.get(input.loci.get(h).populations.get(i).samovaGroupID - 1).populations.size() > 1)
																{
																	double r1 = Math.random();
																	if (r1 < R) 
																		{	
																			R = r1;
																			randomEdge[0] = i; randomEdge[1] = j; 
																		}
																}
														}
												}
											
											// (9.6) the 2 sampled populations located on both sides of this selected edge are identified,
													// and one population chosen at random is assigned to the group of the other population:
											for (int i = 0; i < groups.size(); i++)
												{
													if (groups.get(i).groupID == input.loci.get(h).populations.get(randomEdge[1]).samovaGroupID)
														{
															groups.get(i).populations.add(input.loci.get(h).populations.get(randomEdge[0]));
														}
													if (groups.get(i).groupID == input.loci.get(h).populations.get(randomEdge[0]).samovaGroupID)
														{
															for (int j = 0; j < groups.get(i).populations.size(); j++)
																{
																	if (groups.get(i).populations.get(j).popID == input.loci.get(h).populations.get(randomEdge[0]).popID)
																		{
																			groups.get(i).populations.remove(j);
																		}
																}
														}
												}
									
											// (9.7) the new Fct value associated with this new partition is computed:
												// newFct = this.groupsFct(input.loci.get(h), groups, nodes);
											Fstats = new double[9];
											Fstats = amova.amovaFstatistics(input.loci.get(h), groups, nodes, false, false);
											newFct = Fstats[2];
											// System.out.println(Fct + "	" + newFct);
											/* for (int i = 0; i < input.loci.get(h).populations.size(); i++)
												{
													System.out.print(input.loci.get(h).populations.get(input.loci.get(h).populations.get(i).popID).samovaGroupID);
												}	System.out.println(); */
											
											double p = 0; double r2 = 0; r2 = Math.random();
											if (newFct < Fct)
												{
													p = Math.exp((newFct-Fct)*Math.pow((((double) numberOfSteps) + 1), 0.9158));
												}
											
											if (newFct > Fct)
												{
													newFctValue = true;
												}
											else if (r2 < p)
												{
													newFctValue = true;
												}
											if (newFctValue == true)
												{
													newFctValue = false;
													if (newFct < Fct)
														{
															// System.out.println(numberOfSteps + " " + ((float) p) + " " + ((float) r2));
														}
													Fct = newFct; // System.out.println("newFct = " + Fct);
													int oldID = input.loci.get(h).populations.get(randomEdge[0]).samovaGroupID;
													input.loci.get(h).populations.get(randomEdge[0]).samovaGroupID = input.loci.get(h).populations.get(randomEdge[1]).samovaGroupID;
													for (int i = 0; i < input.loci.get(h).populations.size(); i++)
														{	
															if (delaunayTriangulation[randomEdge[0]][i] == -1)
																{
																	if (input.loci.get(h).populations.get(i).samovaGroupID == input.loci.get(h).populations.get(randomEdge[1]).samovaGroupID)
																		{
																			delaunayTriangulation[randomEdge[0]][i] = 1;
																			delaunayTriangulation[i][randomEdge[0]] = 1;
																		}
																}
															else if (delaunayTriangulation[randomEdge[0]][i] == 1)
																{
																	if (input.loci.get(h).populations.get(i).samovaGroupID == oldID)
																		{
																			delaunayTriangulation[randomEdge[0]][i] = -1;
																			delaunayTriangulation[i][randomEdge[0]] = -1;
																		}
																}											
														}
												}
											else 
												{
													for (int i = 0; i < groups.size(); i++)
														{
															if (groups.get(i).groupID == input.loci.get(h).populations.get(randomEdge[1]).samovaGroupID)
																{
																	for (int j = 0; j < groups.get(i).populations.size(); j++)
																		{
																			if (groups.get(i).populations.get(j).popID == input.loci.get(h).populations.get(randomEdge[0]).popID)
																				{
																					groups.get(i).populations.remove(j);
																				}
																		}
																}
															if (groups.get(i).groupID == input.loci.get(h).populations.get(randomEdge[0]).samovaGroupID)
																{
																	groups.get(i).populations.add(input.loci.get(h).populations.get(randomEdge[0]));
																}
														}
												}
										}
									if (Fct > finalFct)
										{
											finalFct = 0; finalFct = Fct; // System.out.println(finalFct);
											finalGroups = new ArrayList<Group>(); finalGroups.addAll(groups);
											Double currentFct = new Double(Fct);
											allTheFct.add(currentFct);
										}
									else
										{
											Double currentFct = new Double(Fct);
											allTheFct.add(currentFct);
										}
								}
							
							for (int i = 0; i < finalGroups.size(); i++)
								{
									for (int j = 0; j < finalGroups.get(i).populations.size(); j++)
										{
											finalGroups.get(i).populations.get(j).samovaGroupID = finalGroups.get(i).groupID;
										}
								}
							
							// Assignation of the final ID to each Group (cfr. notes in the Group class):
							int g = 1;
							for (int i = 0; i < input.loci.get(h).populations.size(); i++)
								{
									if (finalGroups.get(input.loci.get(h).populations.get(i).samovaGroupID - 1).finalID == 0
											&& g < (k+1))
										{
											finalGroups.get(input.loci.get(h).populations.get(i).samovaGroupID - 1).finalID = g;
											g = g + 1;
										}
								}
							for (int i = 0; i < input.loci.get(h).populations.size(); i++)
								{
									int groupID = input.loci.get(h).populations.get(i).samovaGroupID;
									finalGs = finalGs + finalGroups.get(groupID-1).finalID + "	";
								}
							results.add("best partition for K = " + k + ":	" + finalGs);
							results.add("     PhiCT: " + "	" + (float) finalFct);
							double[] Fstats = new double[12]; Fstats = amova.amova(h, input.loci.get(h), finalGroups, nodes, 0, simTestThread);
							double Fst = 0; Fst = Fstats[1];
							double Fsc = 0; Fsc = Fstats[0];
							results.add("     PhiST: " + "	" + (float) Fst);
							results.add("     PhiSC: " + "	" + (float) Fsc);
							String allTheFctString = new String();
							for (int i = 0; i < allTheFct.size(); i++)
								{
									allTheFctString = allTheFctString + "	" + (float) allTheFct.get(i).doubleValue();
								}
							results.add("     			Final PhiCT value of each repetition: " + allTheFctString);
							
							simTestThread.run("	 for locus " + (h+1));
							simTestThread.run("End of locus by locus SAMOVA analysis for K = " + k);
							
							/* Construction of the SAMOVA barriers matrix:
							externalMessage.add("	");
							externalMessage.add("Barriers matrix (locus by locus SAMOVA) for locus " + (h+1) + " for K = " + k + " (-1 corresponds to a barrier between two sampled populations): ");
							externalMessage.add("	");
							String popID1 = new String();
							ArrayList<String> popID2 = new ArrayList<String>();
							for (int i = 0; i < input.loci.get(h).populations.size(); i++)
								{
									popID1 = popID1 + input.loci.get(h).populations.get(i).popID + "	";
									popID2.add(String.valueOf(input.loci.get(h).populations.get(i).popID));
								}
							externalMessage.add("	" + popID1);					
							for (int i = 0; i < input.loci.get(h).populations.size(); i++)
								{	
									String line = new String();
									line = (popID2.get(i) + "	");
									for (int j = 0; j < input.loci.get(h).populations.size(); j++)
										{
											line = line + (delaunayTriangulation[i][j] + "	");
										}	externalMessage.add(line);
								} */
						}
					results.add("");
				}
		}
}